// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: GPL-3.0-or-later

use {
    crate::helpers::{block_on, set_if_valid},
    crate::nss_bindings::{gaih_addrtuple, nss_status, HErrno},
    crate::{gethostbyaddr, gethostbyname},
    libc::{hostent, size_t, socklen_t, AF_INET},
    nix::errno::Errno,
    std::os::raw::{c_char, c_int, c_void},
    std::ptr,
};

// -------------- by host ---------------

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyname4_r(
    name: *const c_char,
    pat: *mut *mut gaih_addrtuple,
    buffer: *mut c_char,
    buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
    ttlp: *mut i32,
) -> nss_status {
    let mut args = gethostbyname::Args {
        name,
        family: 0,
        result: gethostbyname::Result::V4(pat),
        buffer,
        buflen,
        errnop,
        h_errnop,
        ttlp,
    };

    match block_on(async { gethostbyname::with(&mut args).await }) {
        Ok(status) => status,
        Err(runtime_error) => {
            log::error!("gethostbyname4_r: {}", runtime_error);
            set_if_valid(args.h_errnop, HErrno::Internal);
            nss_status::NSS_STATUS_TRYAGAIN
        }
    }
}

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyname3_r(
    name: *const c_char,
    af: c_int,
    host: *mut hostent,
    buffer: *mut c_char,
    buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
    ttlp: *mut i32,
    canonp: *mut *mut char,
) -> nss_status {
    let mut args = gethostbyname::Args {
        name,
        family: af,
        result: gethostbyname::Result::V3(host),
        buffer,
        buflen,
        errnop,
        h_errnop,
        ttlp,
    };

    match block_on(async { gethostbyname::with(&mut args).await }) {
        Ok(status) => {
            if !host.is_null() {
                set_if_valid(canonp, (*host).h_name as *mut char);
            }
            status
        }
        Err(runtime_error) => {
            log::error!("gethostbyname3_r: {}", runtime_error);
            set_if_valid(h_errnop, HErrno::Internal);
            nss_status::NSS_STATUS_TRYAGAIN
        }
    }
}

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyname2_r(
    name: *const c_char,
    af: c_int,
    host: *mut hostent,
    buffer: *mut c_char,
    buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
) -> nss_status {
    _nss_malcontent_gethostbyname3_r(
        name,
        af,
        host,
        buffer,
        buflen,
        errnop,
        h_errnop,
        ptr::null_mut(),
        ptr::null_mut(),
    )
}

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyname_r(
    name: *const c_char,
    host: *mut hostent,
    buffer: *mut c_char,
    buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
) -> nss_status {
    _nss_malcontent_gethostbyname3_r(
        name,
        AF_INET,
        host,
        buffer,
        buflen,
        errnop,
        h_errnop,
        ptr::null_mut(),
        ptr::null_mut(),
    )
}

// ----------------- by addr -----------------

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyaddr2_r(
    addr: *const c_void,
    len: socklen_t,
    af: c_int,
    host: *mut hostent,
    buffer: *mut c_char,
    buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
    ttlp: *mut i32,
) -> nss_status {
    set_if_valid(errnop, nix::errno::from_i32(0));
    set_if_valid(h_errnop, HErrno::Success);

    match block_on(async {
        gethostbyaddr::with(addr, len, af, host, buffer, buflen, errnop, h_errnop, ttlp).await
    }) {
        Ok(status) => status,
        Err(runtime_error) => {
            log::error!("gethostbyaddr2_r: {}", runtime_error);
            set_if_valid(h_errnop, HErrno::Internal);
            nss_status::NSS_STATUS_TRYAGAIN
        }
    }
}

#[no_mangle]
pub unsafe extern "C" fn _nss_malcontent_gethostbyaddr_r(
    addr: *const c_void,
    len: socklen_t,
    af: c_int,
    host: *mut hostent,
    buffer: *mut c_char,
    buflen: size_t,
    errnop: *mut Errno,
    h_errnop: *mut HErrno,
) -> nss_status {
    _nss_malcontent_gethostbyaddr2_r(
        addr,
        len,
        af,
        host,
        buffer,
        buflen,
        errnop,
        h_errnop,
        ptr::null_mut(),
    )
}
